{ *** WARNING : Don't try to compile this prog with a version of TP other than
  TP6 : I'm not sure that the file type is the same !!! ( cf the small asm
  part ) }

Program MakeSong;
{ Ce programme cree un fichier 83p a partir d'un fichier chanson (.sng)
  Format : fichier texte
   On : Selectionne l'octave N (0<=n<=7) n  entier
   Tn : Tempo = n secondes ( n reel > 0)
   CDEFGABP : Note ... ( P = pause )
   Ln : Duree en unite de tempo ( n reel > 0)
   + : Note precedente est un diese
   - : Note precedente ent un bemol
   < : Passe a l'octave inferieur
   > : Passe a l'octave superieur
   Mn : n = L, N, ou S
         L : Long   : note jouee sur toute sa dure
         N : Normal : note jouee sur 7/8 de sa duree
         S : Short  : note jouee sur 3/4 de sa duree
}

{ =============================== Directives =============================== }

{$A+} {$B-} {$N+} {$E+} {$F-} {$G+} {$I-} {$O-} {$R-} {$S-} {$V+} {$X+}

{-$DEFINE DEBUG}
{$IFDEF DEBUG]
  {$D+} {$L+}
{$ELSE}
  {$D-} {$L-}
{$ENDIF DEBUG}

{ ======================== Declarations, imports, ... ====================== }

Uses Crt;

{ Les notes... Frequences pour l'octave 0
  Pour les octaves au-dessus, on multiplie par 2 }
Const NotesFreq : Array[0..11] of Real
            = ( 65.406,  69.295,  73.416,  77.781,  82.407,  87.307,
                92.499,  97.999, 103.826, 110.000, 116.541, 123.471);
      HeaderSize = $48;

      DLoopCycles = 26;
      SoundLoopCycles = 103;
      TISpeed = 6000000;

Type TNoteType = (Long, Normal, Short);



Var Fin       : Text;       { Handle des fichier d'entree et de sortie }
    Fout      : File of Byte;
    NameIn,
    NameOut   : String;     { Nom des fichiers d'entree et de sortie }
    ProgName  : String[8];  { Nom du prog TI-83 }
    I         : Byte;
    Line      : Word;       { Compteur de lignes }
    OutBuffer : Array[0..32767] of Byte; { Plus rapide  enregistrer }
    OutLength : Word;       { Taille du fichier de sortie }
    Header    : String[HeaderSize]; { En-tete fichier TI-83 }

    Octave    : Byte; { octave en cours }
    Tempo     : Real; { Tempo }
    NoteLen   : Real; { Duree de la note }
    NoteType  : TNoteType; { Type de note }
    Error     : Boolean; { A 1 si au moins une erreur }

    CheckSum  : Word;
    ChkIndex  : Word;

{ ========================== Fin des declarations ========================== }

{ ==================== Debut de la procedure ( 1 seule ) =================== }

Procedure TranslateLine;
{ Lit une ligne du fichier, et l'interprete }
Var LineBuf   : String; { Buffer de lecture }
    CurrentCh : Char; { Caractre trait actuellement }
    Index     : Byte; { Index dans la chaine }
    Value     : Real;

  Function IsMath(Ch : Char) : Boolean;
  { Renvoie True si Ch est un element mathematique }
  Const MathElements = ['0'..'9', '.'];
  Begin
    IsMath := Ch in MathElements;
  End;

  Function ReadValue : Boolean;
  Var ValueStr : String;
      Start    : Byte;
      Erreur   : Integer;
  Begin
    Start := Index;
    Inc(Index);

    While (Index <= Length(LineBuf)) AND IsMath(LineBuf[Index])
      Do Inc(Index);
    Dec(Index);

    If (Index=Start) OR (Index > Length(LineBuf)) Then Begin
      ReadValue := False;
      Exit;
    End;

    ValueStr := Copy(LineBuf, Start+1,Index-Start);
    Val(ValueStr, Value, Erreur);
    ReadValue := (Erreur = 0);
  End;

  Procedure SyntaxError;
  Begin
    WriteLn(NameIn,' Line ',Line, ' : Syntax error at pos ', Index);
    Error := True;
  End;

  Procedure RangeError;
  Begin
    WriteLn(NameIn,' Line ',Line, ' : Range error');
    Error := True;
  End;

  Procedure PutSound(Frequency, Duration : Real);
  Var HL, DE : LongInt;
      RealF  : Real;
  Begin
    HL := Round((TISpeed/2/Frequency-SoundLoopCycles)/DLoopCycles);
    RealF := TISpeed/(2*(DLoopCycles*HL+103));
    DE := Round((2*Duration*Frequency));

  {$IFDEF DEBUG}
    WriteLn('Freq= ', Frequency, '; Duration= ', Duration, '; HL= ',HL, '; DE= ',DE);
  {$ENDIF DEBUG}

    If DE = 0 Then Begin
      { Si le tempo est trop bas, il est possible qu'on aie une duree nulle .. }
      WriteLn(NameIn,' Line ',Line, ' : Internal error : Duration = 0 at pos ', Index);
      Error := True;
      Exit;
    End;

    If HL > 65535 Then Begin
      { N'est cense ne jamais arriver, mais on ne sait jamais... }
      WriteLn(NameIn,' Line ',Line, ' : Internal error : Freq too low at pos ', Index);
      Error := True;
      Exit;
    End;

    { Duree trop longue pour etre codee sur 1 word ? }
    While (DE > 65535) Do Begin
    { Alors ecrit autant de fois qu'il le faut ! }
      Word(Addr(OutBuffer[OutLength])^) := Word(HL);
      INC(OutLength,2);
      Word(Addr(OutBuffer[OutLength])^) := 65535;
      INC(OutLength,2);
      DE := DE - 65535;
    End;

    { Ecrit le son ... }
    Word(Addr(OutBuffer[OutLength])^) := Word(HL);
    INC(OutLength,2);
    Word(Addr(OutBuffer[OutLength])^) := Word(DE);
    INC(OutLength,2);
  End;

  Procedure PutSilence(Duration : Real);
  Begin
    PutSound(TISpeed/2/(DLoopCycles+SoundLoopCycles), Duration);
  End;

  Procedure GetNote;
  { Lit une note et convertit }
  Const NotesListe : String = 'C D EF G A B';
  Var Note : Byte;
      Freq : Real;
      Duration : Real;

  Begin
    Note := Pos(CurrentCh, NotesListe) - 1;

    If (Index < Length(LineBuf)) AND (LineBuf[Index+1] in ['+','-']) Then Begin
      Inc(Index);
      Note := Note - 3 + 2 * Pos(LineBuf[Index], '-+');
      { B+, ou A-, ou ... }
      If (Note > 11) or (Note < 0) Then Begin
        RangeError;
        Exit;
      End;
    End;

    Freq := NotesFreq[Note] * ((2 SHL Octave) DIV 2) ;
    Duration := Tempo*NoteLen;
    Case NoteType of
      Normal: Begin
          PutSound(Freq, Duration*(7/8));
          PutSilence(Duration/8);
              End;

    Long:   PutSound(Freq, Duration);

    Short:  Begin
          PutSound(Freq, Duration*(3/4));
          PutSilence(Duration/4);
            End;
    End;
  End;

  Procedure GetPause;
  { Met une pause dans le fichier de sortie }
  Begin
    PutSilence(Tempo*NoteLen);
  End;


Begin
  Inc(Line);
  ReadLn(Fin, LineBuf);

{$IFDEF DEBUG}
  WriteLn('Ligne ', Line, ' : ', LineBuf);
{$ENDIF DEBUG}

  For Index := 1 to Length(LineBuf) Do Begin
    CurrentCh := UpCase(LineBuf[Index]);

    Case CurrentCh Of
     { Espace : on saute }
      ' ': ;

     { Commentaire : on passe  la ligne suivante }
      ';': Index := Length(LineBuf);

     { Note }
      'A'..'G' : GetNote;
      'P' : GetPause;

     { L : on change la duree }
      'L' : If Not ReadValue Then SyntaxError Else NoteLen := Value;

      '<' : Dec(Octave);
      '>' : Inc(Octave);

      { M : on change le type }
       'M' : If Index < Length(LineBuf) Then Begin
               Inc(Index);
               case UpCase(LineBuf[Index]) of
                'S' : NoteType := Short;
                'N' : NoteType := Normal;
                'L' : NoteType := Long;
               Else
                RangeError;
               End;
             End Else RangeError;

     { O : on change d'octave }
      'O' : If  not ReadValue Then SyntaxError
            Else If (Value-Int(Value)=0)
              Then Octave := Trunc(Value) Else RangeError;

     { T : on change le tempo }
      'T' : If Not ReadValue Then SyntaxError Else Tempo := Value;
    Else
      SyntaxError;
    End;

  If (Octave < 0) OR (Octave > 7) OR (NoteLen <= 0) OR (Tempo <= 0) then RangeError;

{$IFDEF DEBUG}
  WriteLn('T=', Tempo, ' O=', Octave, ' L=', NoteLen, ' M=', Byte(NoteType));
  ReadKey;
{$ENDIF DEBUG}

  End;
End;


Procedure CloseAll; Far;
Begin
  WriteLn;
  Close(Fin);
  Close(Fout);
End;

Begin
{ Initialisation des diverses variables }
  ExitProc := Addr(CloseAll);
  OutLength := HeaderSize;            
  Header := '**TI83**'+#$1A+#$0A+#0+'Song made with MAKESONG for Song Player...'
            +  #0+#0  +#$0B+#0+  #0+#0  +#6;
  { On remplacera les #0+#0 par la taille - $37 (@ $35)
                                          - $46 (@ $39, $44)
                                           -$48 (@ $46) }
  Move(Header[1], OutBuffer[0], HeaderSize);
  Line := 0;
  Error := False;
  For I := 1 to 8 Do
    ProgName[I] := #0;

  Octave  := 3;
  Tempo   := 1/2;
  NoteLen := 1;
  NoteType := Normal;

{ Texte }
  WriteLn ('');
  WriteLn ('*** MakeSong for Song Player V1.0 ***');
  WriteLn ('By Dhordain Florent (jldhordain@nordnet.fr)');
  WriteLn ('');

{ Lecture ligne de commande }
  If ParamCount < 1 Then Begin
    WriteLn ('  Syntax : MAKESONG INPUT [NAME]');
    WriteLn ('  Give filename without extension');
    WriteLn ('  If name is omited, don''t give path');
    Halt(1);
  End;
  NameIn := ParamStr(1)+ '.SNG';
  NameOut := ParamStr(1)+ '.83P';
  If ParamCount > 1 Then
    ProgName := ParamStr(2)
  Else
    ProgName := ParamStr(1);

  For I := 1 to Length(ProgName) Do
    ProgName[I] := UpCase(ProgName[I]);


{ Ouvre le fichier d'entree }
  Assign (Fin, NameIn);
  Reset (Fin);
  If IOResult <> 0 Then Begin
    WriteLn('ERR: Unable to open ',NameIn);
    Halt(1);
  End;

{ Infos }
  WriteLn (' Input file = ',NameIn);
  WriteLn (' Outut file = ',NameOut);
  WriteLn (' Song Name = ',ProgName);

  { Ecrit le nom du prog }
  Move(ProgName[1], OutBuffer[$3C], 8);

  While not EOF(Fin) DO
    TranslateLine;

  Close(Fin);

  { Ecrit les 4 #0 de fin }
  LongInt(Addr(OutBuffer[OutLength])^) := 0;
  INC(OutLength, 4);

{ Le fichier est converti dans OutBuffer }
{ Cree fichier de sortie }
  If Error Then Begin
    WriteLn('There were errors... Sorry.');
    Halt(1);
  End;
  { On ecrit la taille - $37 (@ $35)
                       - $46 (@ $39, $44)
                       - $48 (@ $46) }

  Word(Addr(OutBuffer[$35])^) := OutLength - $37;
  Word(Addr(OutBuffer[$39])^) := OutLength - $46;
  Word(Addr(OutBuffer[$44])^) := OutLength - $46;
  Word(Addr(OutBuffer[$46])^) := OutLength - $48;

  { Calcule la somme de controle }
  CheckSum := 0;
  For ChkIndex := $37 to OutLength-1 Do
    Inc(CheckSum, OutBuffer[ChkIndex]);

  Word(Addr(OutBuffer[OutLength])^) := CheckSum;
  Inc (OutLength, 2);

  Assign (FOut, NameOut);
  ReWrite (FOut);
  If IOResult <> 0 Then Begin
    WriteLn('ERR: Unable to create ',NameOut);
    Halt(1);
  End;
  { Ecrit le bloc d'un seul coup }
  { *** WARNING : I'm not sure that this part of code works with another version
    of TP6... Be very careful if you run it with something else than TP6...
    I couldn't be in any cause responsible of your HARD DISK crash... }
  Asm
    MOV BX, word ptr [FOut]              { BX = Handle }
    MOV CX, [OutLength]
    MOV AH, 040h
    MOV DX, Offset OutBuffer
    INT 021h
    MOV AX, 0
    JNC @fileOK
    MOV AX, 1
  @fileOK:
    MOV [Error], AL
  End;
  Close(Fout);
  IF Error
    Then WriteLn('Error writing ',NameOut)
    Else WriteLn(ProgName, ' is ready to be a great succes !');
End.
